/**********************************************************************************************
*
*	PT100 Box
*	V1.0  27.1.2008
*	(C) 2008 by Lenzen
*
*	Befehle:
*		CM	=  Contiunes Mode  (Fortlaufende Messung der Selektierten Kanaele )
*		ST  =  Stop Continues  ( Stoppt die fortlaufende Messung)
*		A1  =  AutoCalibration (Auto Kalibrierung ein)
*		A0  =                  (Auto Kalibrierung aus)
*		Mx  =  Kanaele auswaehlen
*		Rx  =  Einzelene Kanaele einmal auslesen
*
**********************************************************************************************/


#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include <reg51.h>                /* special function register declarations   */

#define uchar	unsigned char
#define word	unsigned int
#define LF '\n'
#define CR '\r'

extern	int		DualSlope(unsigned char);
extern	void	InitSerial(void);
extern	uchar	GetChar(void);
extern	void	PutChar(uchar);
extern	uchar	GetCharCount(void);

void	TempMeasure(uchar);
void	Strout(char*);
void	SendTemp(uchar);

sbit	LED = 0x91;
float	fTempVal[4];

void	main(void)
{
	bit		AutoCalFlag,RunFlag;
	char	ch,ch1,ch2;
	uchar	ucChan,ucChannels,n;
	float	fa,fb;


	InitSerial();				/* Serielle Schnittstelle auf 9600 Bd initialisieren */
	PT0 = 1;					/* Timer hat hoechste Pioritaet */
	EA = 1;						/* Interrupt freigeben */
	AutoCalFlag = 0;			/* Standartmaessig ohne Auto Calibration */
	RunFlag = 0;				/* Einzelabfrage Mode */
	fa = 1.0;					/* Kalibrationsfaktoren */
	fb = 0.0;

	ch1 = ch2 = 0;
	ucChannels = 0;				/* keine kanaele selectiert */
	n = 0;

	for(;;) // Forever
	{
		// Calibration Mode eingeschaltet ?
		if (AutoCalFlag)
		{
			TempMeasure(0);								// Werte fuer 0 Grad ermitteln
			TempMeasure(1);								// Werte fuer 100 Grad ermitteln
			fa = 100.0 / (fTempVal[1] - fTempVal[0]);	// werte ausrechnen
			fb = 100.0 - fTempVal[1] * fa;
		}
		// Fortlaufende Messung eingeschaltet ?
		if(RunFlag)
		{
			if (ucChannels & (1 << n))
 			{
				TempMeasure(n);
				if(AutoCalFlag)							// muessen die werte korrigiert werden ?
					fTempVal[n] = fTempVal[n] * fa + fb;
				SendTemp(n);
				if(n < 3) PutChar(' ');
			}
			if(n == 3) Strout("\r\n");
			n = (n+1) & 0x03;							// modulo hochzaehlen
		}


		if(GetCharCount())		/* zeichen von schnittstelle ? */
		{
			ch = GetChar();		/* zeichen einlesen */
			
			if (ch == CR)		/* auf CR warten */
			{
				switch(ch1){
				case 'R':		/* Read channel */
					RunFlag = 0;
					ucChan = ch2 - '0';
					if (ucChan <= 3)
					{
						TempMeasure(ucChan);
						if(AutoCalFlag)				// muesen die werte korrigiert werden ?
							fTempVal[ucChan] = fTempVal[ucChan] * fa + fb;
						SendTemp(ucChan);
						Strout("\r\n");
					}
					break;

				case 'S':		/* Stop continues Mode */
					if(ch2 == 'T') RunFlag = 0;
					break;

				case 'C':		/* Start continues Mode */
					if(ch2 == 'M') RunFlag = 1;
					break;

				case 'M':		/* Channel set for continues Mode */
					if(isxdigit(ch2))
					{
						if(isdigit(ch2)) ucChannels = ch2 - '0';
						else			 ucChannels = ch2 - 'A' + 10;
					}
					break;
				case 'A':		/* Calibration mode */
					if(ch2 == '0') AutoCalFlag = 0;
					else
					if(ch2 == '1') AutoCalFlag = 1;
					break;
				} // switch
			} // if (ch = '\r')
			else
			{	// kein CR und kein LF dann speichern
				if(ch != LF)
				{
					ch1 = ch2;		// zeichen speichern
					ch2 = toupper(ch);
				}
			} //else
		} // if (GetCharCount())

	} // for ever
}
/*******************************************************************************************
*
*	void	SendTemp(uchar ucChan)
*	Temperatur ausgeben
*
********************************************************************************************/
void	SendTemp(uchar ucChan)
{
	char	szStr[8];
	LED = 0;
	sprintf(szStr,"%6.2f",fTempVal[ucChan]);
	Strout(szStr);
	LED = 1;
}
/*******************************************************************************************
*
*	void	Strout(char *szStr)
*	String ausgeben
*
********************************************************************************************/
void	Strout(char * szStr)
{
	while(*szStr)
		PutChar(*szStr++);
}
/*******************************************************************************************
*
*	void	TempMeasure(uchar	ucChan)
*	Temperatur Messung vornehmen
*	Kanalnummer in ucChan ( 0..3 )
*	Ergebnis in fTempVal
*
********************************************************************************************/
#define	a 3.9083E-3					/* Linearisierungs konstanten */
#define b -5.8026E-7
#define Ik 0.001					/* Mess Strom */
#define Rs 84.27					/* Widerstand am unteren Messbereichsende hier -40 */
#define k 23.7						/* Verstaerkungsfaktor */
#define k1 (Rs*Ik*k)				/* Verrechnungs faktor */
#define k2 (100.0*k*Ik)				/* Verrechnungs faktor */
#define Step 0.00012207				/* AD Wandler einheit */

void	TempMeasure(uchar	ucChan)
{
	int		iAutoZeroVal;
	int		iTMess;
	uchar	n;
	float	fV;

	// Messkanal setzen
	P1 = (P1 & 0x9f) | (ucChan << 5) & 0x60;

	// Auto zero messung vornehmen
	for(n=0;n<5;n++)
		iAutoZeroVal = DualSlope(0);
	// Temperaturmessung vornehmen
	iTMess = DualSlope(1) - iAutoZeroVal;
	// Messwert linearisieren und speichern
	fV = ((float)iTMess*Step + k1) / k2 - 1.0;
	fTempVal[ucChan] = -sqrt(fV/b + (a/(2*b)) * (a/(2*b)) ) - a/(2*b);
}


